// A plugin to trigger events via a spoken keyword, or short phrase.
// Utilises Unity's KeywordRecognizer which listens to speech input and compares the spoken
// keyword, or words, to a predefined list before triggering the associated event(s).

// ---------------------------- CREDITS -------------------------------
// Almadiel for the original VoiceTrigger plugin on which this is based.
// MacGruber/Acidbubbles for the custom TriggerHandler.
// Acidbubbles for the custom text input field.
// MeshedVR for the Head Up Display example.
// --------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Windows.Speech;
using SimpleJSON;
using AssetBundles;

namespace VeeRifter {
	public class VoiceControl : MVRScript {
		
		protected JSONStorableBool enableLog;
		protected JSONStorableBool enableFocusWarning;		
		protected JSONStorableBool enableHUD;		
		protected UIDynamicButton sharedButton;	
		protected UIDynamicButton cancelButton;			
    	protected JSONStorableStringChooser editTrigger;
    	protected JSONStorableStringChooser renameTrigger;
    	protected JSONStorableStringChooser deleteTrigger;		
    	protected JSONStorableStringChooser setConfidence;
		protected JSONStorableString triggerNameText;
		protected UIDynamicTextField triggerNameTextInput;		
		protected Canvas popupCanvas;
		protected Text popupText;
		protected Text inputTitleText;
		protected InputField inputField;
		
        private const string pluginName = "VoiceControl";
		private const string pluginAuthor = "VeeRifter";
		private const string pluginVersion = "v1.5";
		private const string pluginDate = "27-10-21";	
		private const string aboutText = "\n<b><size=40><color=brown>" + pluginName + " " + pluginVersion + "</color></size></b>\n<i>" + pluginAuthor + "\n" + pluginDate + "</i>";
		private const string stringChooserDefault = "Select";	
		private const string titleTextDefault = "Enter New Trigger Name...";
		private const string popupDefault = "VAM window does not have focus\nClick anywhere inside VAM's window to enable Voice Control";		
		private string[] commandArray;	
		private KeywordRecognizer recognizer = null;
		private static readonly List<string> confidenceList = new List<string>() {"Low","Medium","High"};
		private ConfidenceLevel confidence = ConfidenceLevel.Medium;
		private VoiceTrigger voiceTrigger;		
		private List<VoiceTrigger> voiceTriggerList;
		private List<string> triggerNameList;
		private List<string> commandList;
    	private List<string> micDeviceList;		
		private int listIndex;		
		private bool renaming = false;
		private bool deleting = false;
		private bool addOrEdit = false;
		private bool showHUD = false;
		private bool firstPass = true;		
    	private const float showHUDduration = 3.0f;
    	private float showHUDtimer = 0.0f;
		
		public override void Init() {
			try {
				pluginLabelJSON.val = pluginName + " " + pluginVersion + " - " + pluginAuthor + " - " + pluginDate;
				
				// Load trigger handler screen assets
				SimpleTriggerHandler.LoadAssets();
				
				// Add the rename listener
				SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
				
				//Create the lists
				micDeviceList = new List<string>();				
				commandList  = new List<string>();
				voiceTriggerList = new List<VoiceTrigger>();		
				triggerNameList = new List<string>();
				triggerNameList.Add(stringChooserDefault);
				
				// Create UI left side				
				UIDynamic leftSpacer1 = CreateSpacer(false);
				leftSpacer1.height = 15f;
				
				JSONStorableString about = new JSONStorableString("About", aboutText);
				UIDynamicTextField dtfAbout = CreateTextField(about, false);
				dtfAbout.UItext.alignment = TextAnchor.MiddleCenter;
				dtfAbout.UItext.supportRichText = true;
				dtfAbout.UItext.fontSize = 30;		
				dtfAbout.UItext.color = Color.black;
				dtfAbout.height = 185;

        		foreach (string device in Microphone.devices)
				{
	            	micDeviceList.Add (device);
				}
				
				string defaultMic = "<b><size=24><color=blue>Active Microphone</color></size></b><i>\n";
				if (micDeviceList.Count > 0)
				{
					//Windows default mic is always first in list
					defaultMic += micDeviceList[0];
				}
				else
				{
					defaultMic += "None";
				}
				defaultMic += "</i>";
				
				JSONStorableString activeMic = new JSONStorableString("Active Mic", defaultMic);
				UIDynamicTextField dtfActiveMic = CreateTextField(activeMic, false);
				dtfActiveMic.UItext.alignment = TextAnchor.MiddleCenter;
				dtfActiveMic.UItext.fontSize = 20;		
				dtfActiveMic.UItext.color = Color.black;
				dtfActiveMic.UItext.supportRichText = true;
				dtfActiveMic.height = 85;

                enableLog = new JSONStorableBool("Log Plugin Messages", true);
                enableLog.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(enableLog);
                CreateToggle(enableLog, false);
				
                enableFocusWarning = new JSONStorableBool("HUD - Window Focus Warning", true);
                enableFocusWarning.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(enableFocusWarning);
                CreateToggle(enableFocusWarning, false);				

                enableHUD = new JSONStorableBool("HUD - Recognizer Feedback", true);
                enableHUD.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(enableHUD);
                CreateToggle(enableHUD, false);

				UIDynamic leftSpacer3 = CreateSpacer(false);
				leftSpacer3.height = 20f;

			    setConfidence = new JSONStorableStringChooser("Confidence Level", confidenceList, "Medium", "Required Confidence Level", setConfidenceLevel);
 		      	RegisterStringChooser(setConfidence);
		        CreateScrollablePopup(setConfidence, false).popupPanelHeight = 220f;
				
				// Create UI right side
				triggerNameText = new JSONStorableString("triggerNameText","");				
				triggerNameTextInput = CreateTextInput(triggerNameText, titleTextDefault, true);

				UIDynamic rightSpacer1 = CreateSpacer(true);
				rightSpacer1.height = 0f;
				
				sharedButton = CreateButton("Add Trigger", true);
				sharedButton.buttonText.color = Color.black;
				sharedButton.buttonColor = Color.green;				
				sharedButton.button.onClick.AddListener(doTrigger);
				
				cancelButton = CreateButton("Cancel", true);
				cancelButton.buttonText.color = Color.white;
				cancelButton.buttonColor = Color.blue;				
				cancelButton.button.onClick.AddListener(resetUI);				

				string availMics = "<b><size=24><color=blue>Available Microphones</color></size></b><i>\n";
				foreach (string mic in micDeviceList)
				{
					availMics += mic;
					availMics += "\n";
				}
				if (micDeviceList.Count == 0)
				{
					availMics += "None";
				}
				availMics += "</i>";
				
				JSONStorableString availableMics = new JSONStorableString("Available Mics", availMics);
				UIDynamicTextField dtfAvailableMics = CreateTextField(availableMics, true);
				dtfAvailableMics.UItext.alignment = TextAnchor.MiddleCenter;
				dtfAvailableMics.UItext.fontSize = 20;		
				dtfAvailableMics.UItext.color = Color.black;
				dtfAvailableMics.UItext.supportRichText = true;
				dtfAvailableMics.height = 85;			
				
		        editTrigger = new JSONStorableStringChooser("Edit Trigger", triggerNameList, stringChooserDefault, "Edit   Trigger", TriggerEdit);
				CreateScrollablePopup(editTrigger, true).popupPanelHeight = 780f;			
				editTrigger.popup.labelTextColor = Color.blue;				
				
		        renameTrigger = new JSONStorableStringChooser("Rename Trigger", triggerNameList, stringChooserDefault, "Rename Trigger", TriggerRename);
		        CreateScrollablePopup(renameTrigger, true).popupPanelHeight = 665f;
				renameTrigger.popup.labelTextColor = Color.yellow;
				
		        deleteTrigger = new JSONStorableStringChooser("Delete Trigger", triggerNameList, stringChooserDefault, "Delete Trigger", TriggerDelete);
		        CreateScrollablePopup(deleteTrigger, true).popupPanelHeight = 550f;	
				deleteTrigger.popup.labelTextColor = Color.red;	
				
				//Create the head-up display
				initHUD();
			}
			catch (Exception e) {
				SuperController.LogError("Exception caught in Init(): " + e);
			}
		}

		void Update()
		{
			try
			{
				bool active = false;
				for (int index = 0; index < voiceTriggerList.Count; ++index)
				{
					voiceTriggerList[index].Update();
					// Monitor each trigger's action panel for activity					
					if (voiceTriggerList[index].actionsPanelActive())
					{
						addOrEdit = true;
						active = true;					
					}						
				}
				
				// Ensure all action panels are closed before updating keyword list and(re)starting recognizer.			
				if (addOrEdit && !active)
				{
					updateCommandList();
					startRecognizer();
				}
				if (showHUD)
				{
					popupText.enabled = true;
					showHUDtimer += Time.deltaTime;
	       			if (showHUDtimer > showHUDduration)
	        		{
						showHUD = false;
            			showHUDtimer = 0;
						popupText.text = popupDefault;
						popupText.enabled = false;					
        			}
				}					
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught in Update(): " + e);
			}
		}

		void OnEnable()
		{
			startRecognizer();			
		}
		
		void OnDisable()
		{
			removeRecognizer();
			logMessage("VoiceControl Plugin is disabled.");		
		}
		
		void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;			
			removeRecognizer();
			if (popupCanvas != null)
				Destroy(popupCanvas.gameObject);
			
			foreach(VoiceTrigger vt in voiceTriggerList)				
				vt.Remove();				
		}

		void OnApplicationFocus(bool hasFocus)
		{
			popupText.enabled = !hasFocus && enableFocusWarning.val;
			if (!hasFocus)
				SuperController.LogMessage(popupDefault);						
		}

    	void OnApplicationPause(bool isPaused)
    	{			
        	popupText.enabled = isPaused && enableFocusWarning.val;
			if (isPaused)			
				SuperController.LogMessage(popupDefault);			
    	}
			
		private void OnAtomRename(string oldid, string newid)
		{			
			foreach(VoiceTrigger vt in voiceTriggerList) {
				vt.SyncAtomNames();
			}
		}

		private void setConfidenceLevel(string level)
		{	// Any keyword recognized, but below the set minimum confidence level, is ignored. 		
			switch(level)
			{
				case "Low":		
					confidence = ConfidenceLevel.Low;
					break;
				case "Medium":		
					confidence = ConfidenceLevel.Medium;
					break;
				case "High":		
					confidence = ConfidenceLevel.High;
					break;
			}	
			startRecognizer();
		}

		private void doTrigger()
		{	
			string triggerName = "";
			
			if(triggerNameText.val == "")
			{	
				inputTitleText.text = "Name Cannot Be Empty...";		
				inputTitleText.color = Color.red;
				return;
			}

			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{	
				triggerName = voiceTriggerList[index].Name;
						
				if (triggerNameText.val == triggerName && !deleting)
				{	
					inputTitleText.color = Color.red;					
					if (renaming)
					{
						inputTitleText.text = "Name identical or in use. Retry or Cancel...";					
					}
					else
					{
						inputTitleText.text = "Name in use. Retry or Cancel...";					
					}
					return;					
				}
			}
			
			if (renaming)
			{			
				triggerName = voiceTriggerList[listIndex].Name;				
				voiceTriggerList[listIndex].Name = triggerNameText.val;
				triggerNameList[listIndex + 1] = triggerNameText.val;							
				logMessage("Renamed Voice Trigger \"" + triggerName + "\" to \"" + voiceTriggerList[listIndex].Name + "\"");								
			}					
			else if (deleting)
			{
				triggerName = voiceTriggerList[listIndex].Name;
				voiceTriggerList.RemoveAt(listIndex);
				triggerNameList.RemoveAt(listIndex + 1);					
            	logMessage("Deleted Voice Trigger \"" + triggerName + "\"");
				updateCommandList();
				startRecognizer();				
			}
			else //Must be adding
			{		
				int count = voiceTriggerList.Count;
				voiceTrigger = new VoiceTrigger(this, triggerNameText.val);		
				voiceTriggerList.Add(voiceTrigger);
				triggerNameList.Add(triggerNameText.val);	
            	logMessage("Added new Voice Trigger \"" + voiceTrigger.Name + "\"");			
				voiceTriggerList[count].OpenPanelActionStart();	
			}		
			resetUI();					
		}

		private void TriggerEdit(string selected)
		{	
			if (editTrigger.val == stringChooserDefault)
				return;
			
			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{
				string triggerName = voiceTriggerList[index].Name;				
				if (selected == triggerName)
				{
					voiceTriggerList[index].OpenPanelActionStart();					
				}					
			}		
			resetUI();
		}
		
		private void TriggerRename(string selected)
		{	
			if (renameTrigger.val == stringChooserDefault)
				return;
						
			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{
				string triggerName = voiceTriggerList[index].Name;				
				if (selected == triggerName)
				{
					renaming = true;
					inputTitleText.text = "Enter New Name...";		
					inputTitleText.color = Color.yellow;
					triggerNameText.val = triggerName;
					sharedButton.label = "Confirm Rename";
					sharedButton.buttonText.color = Color.black;
					sharedButton.buttonColor = Color.yellow;											
					listIndex = index;
				}					
			}				
		}
		
		private void TriggerDelete(string selected)
		{	
			if (deleteTrigger.val == stringChooserDefault)
				return;
			
			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{	
				string triggerName = voiceTriggerList[index].Name;				
				if (selected == triggerName)
				{
					deleting = true;
					inputTitleText.text = "Delete Trigger...";		
					inputTitleText.color = Color.red;
					triggerNameText.val = triggerName;
					sharedButton.label = "Confirm Delete";
					sharedButton.buttonText.color = Color.white;
					sharedButton.buttonColor = Color.red;						
					listIndex = index;
					inputField.enabled = false;		
				}					
			}			
		}

		private void resetUI()
		{
			editTrigger.choices = null;					
			editTrigger.choices = triggerNameList;
			renameTrigger.choices = null;					
			renameTrigger.choices = triggerNameList;			
			deleteTrigger.choices = null;
			deleteTrigger.choices = triggerNameList;
			renameTrigger.val = stringChooserDefault;		
			deleteTrigger.val = stringChooserDefault;								
			editTrigger.val = stringChooserDefault;		
			inputTitleText.text = titleTextDefault;			
			inputTitleText.color = Color.green;
			inputField.enabled = true;	
			triggerNameText.val	= "";
			sharedButton.label = "Add Trigger";
			sharedButton.buttonText.color = Color.black;
			sharedButton.buttonColor = Color.green;				
			deleting = false;
			renaming = false;			
		}
		
		private void updateCommandList()
		{	
			if (commandList.Count > 0)
			{
				logMessage("Updating the command keyword list.");				
				commandList.Clear();
			}

			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{			
				JSONClass json = voiceTriggerList[index].GetJSON();				
		        foreach (JSONNode node in (JSONArray)json["startActions"])
 				{
					string command = (string)node["name"];				 
		            if(command != null && command.Length > 0)
					{
	 	               	commandList.Add(command);
						logMessage("Command \"" + command + "\" added from \"" + voiceTriggerList[index].Name + "\"");
					}					
				}
			}
		addOrEdit = false;			
		}	

		private void startRecognizer()
		{				
		    if (commandList.Count > 0)
		    {
				removeRecognizer();				
				commandArray = commandList.ToArray();				
		    	recognizer = new KeywordRecognizer(commandArray, confidence);				
 		    	recognizer.OnPhraseRecognized += phraseRecognized;				
		    	if (recognizer != null)
				{
 		    		recognizer.Start();				
					logMessage("Recognizer started, using minimum confidence level of \"" + confidence + "\"");
				}
				else
				{
					SuperController.LogError("Unable to create the Recognizer task !");
				}					
 		   	}			
		}	
		
		private void removeRecognizer()
		{
		    if (recognizer != null)
		    {
				if (recognizer.IsRunning)
					recognizer.Stop();
		        recognizer.OnPhraseRecognized -= phraseRecognized;
				recognizer.Dispose();
				recognizer = null;				
		    }
		}
	
		private void phraseRecognized(PhraseRecognizedEventArgs args)
		{
	
			for (int index = 0; index < voiceTriggerList.Count; ++index)
			{
				JSONClass json = voiceTriggerList[index].GetJSON();				
		        foreach (JSONNode node in (JSONArray)json["startActions"])
 				{
					string command = (string)node["name"];
					if(command == args.text)
					{
						//Trigger the associated event
						voiceTriggerList[index].Trigger();
						logMessage("Command \"" + args.text + "\" recognized with " + args.confidence + " confidence.");
						logMessage("Voice Trigger \"" + voiceTriggerList[index].Name + "\" activated.");						
						HUDMessage("Command \"" + args.text + "\" recognized with " + args.confidence + " confidence.");							
					}						
				}
			}
		}
		
		private void logMessage(string message)
		{
			if (!enableLog.val) return; 
			SuperController.LogMessage(message);
		}		
		
		private void HUDMessage(string message)
		{
			if (!enableHUD.val) return; 			
			popupText.text = message;
			showHUD = true;
		}
				
		private void initHUD()
		{		
			GameObject g1 = new GameObject();
			g1.name = "Canvas";				
			popupCanvas = g1.AddComponent<Canvas>();
			popupCanvas.renderMode = RenderMode.WorldSpace;
			CanvasScaler cs = g1.AddComponent<CanvasScaler>();
			cs.scaleFactor = 100.0f;
			cs.dynamicPixelsPerUnit = 1f;
			RectTransform rt = g1.GetComponent<RectTransform>();
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, 500f);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 100f);
			g1.transform.localScale = new Vector3(0.0006f, 0.0006f, 0.0006f);
			g1.transform.localPosition = new Vector3(0.0f, 0.0f, .5f);		
			// anchor popup in front of head
			Transform headCenter = SuperController.singleton.centerCameraTarget.transform;
			rt.SetParent(headCenter, false);

			GameObject g2 = new GameObject();
			g2.name = "popupText";
			g2.transform.parent = g1.transform;
			g2.transform.localScale = Vector3.one;
			g2.transform.localPosition = Vector3.zero;
			g2.transform.localRotation = Quaternion.identity;
			popupText = g2.AddComponent<Text>();
			RectTransform rt2 = g2.GetComponent<RectTransform>();
			rt2.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, 500f);
			rt2.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 100f);
			popupText.alignment = TextAnchor.MiddleCenter;
			popupText.horizontalOverflow = HorizontalWrapMode.Overflow;
			popupText.verticalOverflow = VerticalWrapMode.Overflow;
			Font ArialFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
			popupText.font = ArialFont;
			popupText.fontSize = 24;
			popupText.text = popupDefault;
			popupText.enabled = false;
			popupText.color = Color.white;			
		}


        public UIDynamicTextField CreateTextInput(JSONStorableString jss,string titleText, bool rightSide = false) 
	    {	
            var container = new GameObject();
            container.transform.SetParent(transform, false);
            {
                var rect = container.AddComponent<RectTransform>();
                rect.pivot = new Vector2(0, 1);

                var layout = container.AddComponent<LayoutElement>();
                layout.preferredHeight = 70f;
                layout.flexibleWidth = 1f;
            }

            var textfield = Instantiate(this.manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();

            textfield.gameObject.transform.SetParent(container.transform, false);
            {
                jss.dynamicText = textfield;
                textfield.backgroundColor = Color.white;
                inputField = textfield.gameObject.AddComponent<InputField>();
                inputField.textComponent = textfield.UItext;
				inputField.characterLimit = 45;			
                jss.inputField = inputField;
                Destroy(textfield.GetComponent<LayoutElement>());
                var rect = textfield.GetComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 1);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 1);
                rect.anchoredPosition = new Vector2(0, -30f);
                rect.sizeDelta = new Vector2(0, 50f);
            }

            var title = new GameObject();		
            title.transform.SetParent(container.transform, false);
            {
                var rect = title.AddComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 1);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 1);
                rect.anchoredPosition = new Vector2(0, 0f);
                rect.sizeDelta = new Vector2(0, 30f);
                inputTitleText = title.AddComponent<Text>();
                inputTitleText.font = textfield.UItext.font;
				inputTitleText.fontStyle = FontStyle.Bold;
                inputTitleText.text = titleTextDefault;
                inputTitleText.fontSize = 24;
				inputTitleText.color = Color.green;				
            }
		
			if (rightSide)
				rightUIElements.Add(container.transform);
			else
				leftUIElements.Add(container.transform);
			
            return textfield;
        }

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{

			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				JSONArray vta = new JSONArray();			
				for (int index = 0; index < voiceTriggerList.Count; ++index)
				{
					JSONClass vtc = new JSONClass();
					vtc["Name"] = voiceTriggerList[index].Name;	
					vtc[voiceTriggerList[index].Name] = voiceTriggerList[index].GetJSON(base.subScenePrefix);
					vta.Add("", vtc);					
				}
				jc["VoiceTriggers"] = vta;				
			}
			return jc;
		}
		
		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

			if (firstPass)
			{
				firstPass = false;
				return;
			}
			
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				JSONArray vta = jc["VoiceTriggers"].AsArray;				
				for (int index = 0; index < vta.Count; ++index)
				{				
					JSONClass vtc = vta[index].AsObject;
					voiceTrigger = new VoiceTrigger(this, vtc["Name"]);
					voiceTrigger.RestoreFromJSON(vtc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);		
					voiceTriggerList.Add(voiceTrigger);
					triggerNameList.Add(voiceTrigger.Name);					
				}
				editTrigger.choices = null;					
				editTrigger.choices = triggerNameList;					
				renameTrigger.choices = null;					
				renameTrigger.choices = triggerNameList;					
				deleteTrigger.choices = null;
				deleteTrigger.choices = triggerNameList;
				updateCommandList();				
				string enabled = jc.HasKey("enabled") ? jc["enabled"].Value : null;
				if (enabled != "false")
					startRecognizer();
			}
		}	
	}		
		
	public class SimpleTriggerHandler : TriggerHandler
	{
		public static bool Loaded { get; private set; }

		private static SimpleTriggerHandler myInstance;

		private  RectTransform myTriggerActionsPrefab;
		private RectTransform myTriggerActionMiniPrefab;		
		private RectTransform myTriggerActionDiscretePrefab;
		private RectTransform myTriggerActionTransitionPrefab;

		public static SimpleTriggerHandler Instance {
			get {
				if (myInstance == null)
					myInstance = new SimpleTriggerHandler();
				return myInstance;
			}
		}

		public static void LoadAssets()
		{
			SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
		}

		private IEnumerator LoadAssetsInternal()
		{
			foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
				yield return x;
			foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
				yield return x;
			foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
				yield return x;
			foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
				yield return x;				
			Loaded = true;	
		}

		private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
		{
			AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
			if (request == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
			yield return request;
			GameObject go = request.GetAsset<GameObject>();
			if (go == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
			RectTransform prefab = go.GetComponent<RectTransform>();
			if (prefab == null)
				throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
			assign(prefab);
		}

		void TriggerHandler.RemoveTrigger(Trigger t)
		{	// Cleanup
			if (t.triggerActionsPanel != null)
        		UnityEngine.Object.Destroy(t.triggerActionsPanel.gameObject);			
		}

		void TriggerHandler.DuplicateTrigger(Trigger t)
		{
			throw new NotImplementedException();
		}

		RectTransform TriggerHandler.CreateTriggerActionsUI()
		{
			RectTransform clone = UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			clone.Find("Content/Tab1/Label").GetComponent<Text>().text = "Keyword List";
			clone.Find("Content/Tab1/AddDiscreteStartActionButton/Text").GetComponent<Text>().text = "Add a Keyword";				
			clone.Find("Content/Tab2").gameObject.SetActive(false);
			clone.Find("Content/Tab3").gameObject.SetActive(false);	
			return clone;
		}

		RectTransform TriggerHandler.CreateTriggerActionMiniUI()
		{
			RectTransform clone = UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			clone.Find("NameInputField/Placeholder").GetComponent<Text>().text = "Enter Keyword or Phrase...";			
			clone.Find("Buttons/Open Button/Text").GetComponent<Text>().text = "Action...";
			clone.Find("Buttons/Duplicate Button").gameObject.SetActive(false);		
			return clone;
		}

		RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
		{
			RectTransform clone = UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			clone.Find("NameInputField/Placeholder").GetComponent<Text>().text = "Enter Keyword or Phrase...";
			return clone;			
		}

		RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
		{
			RectTransform clone = UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);	
			return clone;
		}

		void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
		{		
				UnityEngine.Object.Destroy(rt?.gameObject);
		}
	
	}

	public class VoiceTrigger : Trigger
	{				
		public string Name {
			get { return name; }
			set { name = value; myNeedInit = true; }
		}

		public MVRScript Owner {
			get; private set;
		}
		
		private string name;
		private bool myNeedInit = true;
		
		public VoiceTrigger(MVRScript owner, string name)
		{
			Name = name;
			Owner = owner;
			handler = SimpleTriggerHandler.Instance;
		}		

		public bool actionsPanelActive()
		{
			Transform actionsPanel = triggerActionsPanel;
			if (actionsPanel != null)
			{
				
				return actionsPanel.gameObject.activeSelf;
			}
			return false;
		}

		public void OpenPanelActionStart()
		{			
			if (!SimpleTriggerHandler.Loaded)
			{
				SuperController.LogError("VoiceTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
				return;
			}

			triggerActionsParent = Owner.UITransform;
			InitTriggerUI();			
			OpenTriggerActionsPanel();
			
			if (myNeedInit)
			{
				triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;			
				myNeedInit = false;		
			}
		}		

		public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
		{
			if (jc.HasKey(Name))
			{
				JSONClass tc = jc[Name].AsObject;
				if (tc != null)
					base.RestoreFromJSON(tc, subScenePrefix, isMerge);
			}
			else if (setMissingToDefault)
			{
				base.RestoreFromJSON(new JSONClass());
			}
		}
			
		public void Trigger()
		{			
			active = true;
			active = false;
		}		
	}		
}